// BFMix will be a GUI representation of a number of Player Objects
BFMix {

	var <playarray, <mixgui, guichild;
	var starttime, endtime, durstring;
	var mixdur = 0;
	var routeout; // for the Player to route to the global controls
	var server, tempoclock, timestart, timestop, playbutton;
	
	classvar formats, numouts, outvals;
	
	*new {arg server;
		^super.new.init(server);
	}
	
	init {arg serv;
		server = serv ? Server.internal.boot;
		// make this 8 for now... unlimited later?
		playarray = Array.newClear(8);
		this.gui;
//		routeout = server.options.numOutputBusChannels + server.options.numInputBusChannels;
		routeout = 0;
		tempoclock = TempoClock.new;
	}
	
	addTrack {arg track, path;
		var player;
		// create a new player
		player = Player.new(path, outbus: routeout);
		playarray.put(track, player);
		// update the GUI objects
		this.updateTimes(track, 0.0, player.sf.duration, 0.0);
		guichild[track * 15 + 1].string_(player.filename);
		guichild[track * 15 + 13].string_(player.env.asCompileString);
	}
	
	cleartrack {arg track;
		playarray.put(track, nil);
		guichild[track * 15 + 1].string_("");
		guichild[track * 15 + 3].value_(0);	
		guichild[track * 15 + 5].value_(0.7079402);	
		guichild[track * 15 + 6].value_(0);	
		guichild[track * 15 + 8].value_(0);	
		guichild[track * 15 + 10].value_(0);	
		guichild[track * 15 + 11].value_(0);	
		guichild[track * 15 + 13].value_(nil);	
		guichild[track * 15 + 14].lo_(0).hi_(0);
	}
	
	updateTimes {arg track, trackstarttime, trackendtime, mixstart;
		// update mixdur if needed
		(((trackendtime - trackstarttime) + mixstart) > mixdur).if(
			{this.mixdur_((trackendtime - trackstarttime) + mixstart);
				this.endtime_(mixdur.trunc(0.1));
				durstring.string_("Mix dur = "++mixdur.trunc(0.1))});
		guichild[track * 15 + 3].value_(mixstart);
		guichild[track * 15 + 8].value_(trackstarttime);
		guichild[track * 15 + 10].value_(trackendtime);
		guichild[track * 15 + 14]
			.lo_(mixstart / mixdur)
			.hi_((mixstart + (trackendtime - trackstarttime)) / mixdur);
	}
	
	// the global starttime in the mix
	starttime_ {arg newtime;
		starttime.valueAction_(newtime);
	}
	
	// the global end time
	endtime_ {arg newtime;
		endtime.valueAction_(newtime)
	}
	
	// if the mixdur changes, redo the range sliders
	mixdur_ {arg newdur;
		mixdur = newdur;
		playarray.do{arg me, i;
			// if there is a Player there, resize the sliders
			me.notNil.if({
				this.drawTrack(i)})
		}
	}
	
	// this tells you where in the file to start playing
	trackstart {arg track, newstart;
		playarray[track].starttime_(newstart);
		this.updateTimes(
			track, newstart, guichild[track * 15 + 10].value, guichild[track * 15 + 3].value)
	}

	// this tells you where in the file to stop playing
	trackend {arg track, newend;
		playarray[track].endtime_(newend);
		this.updateTimes(
			track, guichild[track * 15 + 8].value, newend, guichild[track * 15 + 3].value)
	}
	
	trackenv {arg track, env;
		playarray[track].env_(env);
		this.updateTimes(track, playarray[track].starttime, playarray[track].endtime,
			guichild[track * 15 + 3].value);
	}
	
	drawTrack {arg track;
		guichild[track * 15 + 14]
			.lo_(guichild[track * 15 + 3].value / mixdur)
			.hi_(guichild[track * 15 + 10].value / mixdur)
	}
	
	playmix {
		timestart.value;
		playarray.do{arg me, i;
			me.notNil.if({
				(starttime.value > guichild[i * 15 + 3].value).if({
					playarray[i].starttime_((guichild[i * 15 + 8].value + 
						(starttime.value - guichild[i * 15 + 3].value)).postln)}, {
					playarray[i].starttime_(guichild[i * 15 + 8].value)});
				me.pausedplay((guichild[i * 15 + 3].value - starttime.value).max(0.0)
				)})};
		tempoclock.sched(endtime.value - starttime.value, {this.stopmix});
	}
	
	stopmix {
		{timestop.value; playbutton.value_(0)}.defer;
		tempoclock.clear;
		playarray.do{arg me, i;
			me.notNil.if({
				me.stop})};		
	}
	 	
	gui {			
		var guiwidth, guiheight, lineheight, itemheight;
		var texttask, clock;
		
		timestart = {
			var epoch, start, minsdisp,time, secs;
			texttask = Task({
				epoch = Main.elapsedTime;
				start = starttime.value;
				loop{
					time = start + (Main.elapsedTime - epoch) ;
					clock.string_(time.asTimeString );
					0.1.wait;
				}
			}, AppClock);
			texttask.play
		};
			
		timestop = {
			var time, secs;
			texttask.stop; 
			time = starttime.value;
			clock.string_(time.asTimeString);
			
		};
					
		mixgui = SCWindow("BFMixer", 
			Rect(150, 0, SCWindow.screenBounds.width - 250, SCWindow.screenBounds.height)).front;
		
		guiwidth = mixgui.bounds.width;
		guiheight = mixgui.bounds.height;
		lineheight = guiheight - 250 * 0.03;
		itemheight = guiheight - 250 * 0.13;
		
		// The tracks
		8.do{arg i;
			// 15 objects per track
			// the load button (track * 15)
			SCButton(mixgui, Rect(guiwidth * 0.02, itemheight * i, guiwidth * 0.09, lineheight))
				.states_([
					["load file", Color.black, Color(0.7, 0.3, 0.3, 0.3)],
					["", Color.black, Color(0.3, 0.7, 0.3, 0.3)]])
				.action_({arg me;
					[{this.cleartrack(i)},{CocoaDialog.getPaths({ arg paths;
						this.addTrack(i, paths[0])}, {"cancelled"})
						}][me.value].value;
					});
						
			// filename (track * 15) + 1
			SCStaticText(mixgui, 
					Rect(guiwidth * 0.02, itemheight * i, guiwidth * 0.09, lineheight))
				.string_("") // this will change to the filename when one is loaded
				.align_(\center);
			
			// static... doesn't change (track * 15) + 2
			SCStaticText(mixgui, 
					Rect(guiwidth * 0.1 + 15, itemheight * i, guiwidth * 0.1, lineheight))
				.string_("Start time in Mix");
			
			// starttime in mix (track * 15) + 3
			SCNumberBox(mixgui, 
					Rect(guiwidth * 0.2 + 10, itemheight * i, guiwidth * 0.03, lineheight))
				.value_(0.0)
				.action_({arg me;
					this.updateTimes(i, guichild[i * 15 + 8].value, 
						guichild[i * 15 + 10].value, me.value)});
			
			// static... doesn't change (track * 15) + 4
			SCStaticText(mixgui, 
					Rect(guiwidth * 0.25, itemheight * i, guiwidth * 0.1, lineheight))
				.string_("Amp (in dB)");
			
			// Amp slider (track * 15) + 5
			SCSlider(mixgui, 
					Rect(guiwidth * 0.33, itemheight * i, guiwidth * 0.3, lineheight))
				.value_(0.7079402)
				.action_({arg me;
					var val;
					val = ControlSpec(-inf, 6, \db).map(me.value);
					guichild[guichild.indexOf(me) + 1].value_(
						val.trunc(0.01));
					guichild[guichild.indexOf(me) + 1].bounds.left_(
						me.bounds.left);
					playarray[i].amp_(val.dbamp);
					});
			
			// amp value (track * 15) + 6
			SCNumberBox(mixgui, 
					Rect(guiwidth * 0.65, itemheight * i, guiwidth * 0.03, lineheight))
				.value_(0.0)
				.action_({arg me;
					var val;
					val = ControlSpec(-inf, 6, \db).unmap(me.value); //map from -inf to 6db
					guichild[guichild.indexOf(me) - 1].value_(
						val.postln);
					});
			
			// static... doesn't change (track * 15) + 7
			SCStaticText(mixgui, 
					Rect(guiwidth * 0.02, itemheight * i + lineheight + 4, guiwidth * 0.1,
						lineheight))
				.string_("Start Time in File =");
				
			// advance into sndfile (track * 15) + 8
			SCNumberBox(mixgui,
					Rect(guiwidth * 0.12, itemheight * i + lineheight + 4, guiwidth * 0.045, 
						lineheight))
				.value_(0.0)
				.action_({arg me;
					this.trackstart(i, me.value)});		
			
			// static... doesn't change (track * 15) + 9
			SCStaticText(mixgui, 
					Rect(guiwidth * 0.18, itemheight * i + lineheight + 4, guiwidth * 0.1, 
						lineheight))
				.string_("End Time in File = ");
			
			// filetime end (track * 15) + 10
			SCNumberBox(mixgui,
					Rect(guiwidth * 0.28, itemheight * i + lineheight + 4, guiwidth * 0.045, 
						lineheight))
				.value_(0.0)
				.action_({arg me;
					this.trackend(i, me.value)});
			
			// output choices (track * 15) + 11
			SCPopUpMenu(mixgui, 
					Rect(guiwidth * 0.85, itemheight * i + lineheight + 4, guiwidth * 0.07, 
						lineheight))
				.items_([
					"straight", "stereo", "uhj", "quad", "hex"])
				.action_({arg me;
					var format;
					format = me.items[me.value];
					((playarray[i].sf.numChannels == 3) || 
						(playarray[i].sf.numChannels == 4)).if({
							playarray[i].formatOut_(format.asSymbol)}, {
							"You can only change the format of a BFormat Signal".warn;
							me.value_(0)});
					});
			
			// static... doesn't change (track * 15) + 12
			SCStaticText(mixgui,
					Rect(guiwidth * 0.33, itemheight * i + lineheight + 4, guiwidth * 0.03, 
						lineheight))
				.string_("Env");
				
			// takes Env args (track * 15) + 13
			SCTextField(mixgui, 
					Rect(guiwidth * 0.36, itemheight * i + lineheight + 4, guiwidth * 0.47, 
						lineheight))
				.value_("nil")
				.action_({arg me; 
					var env;
					me.value.interpret.isKindOf(Env).if({
						env = me.value.interpret;
						this.trackenv(i, env);
						}, {
						"Must be an env object".warn;
						me.value_(playarray[i].env.asCompileString)})
					});
			
			// represents the tracks position in mix graphically
			// (track * 15) + 14
			SCRangeSlider(mixgui, 
					Rect(guiwidth * 0.02, itemheight * i + (lineheight * 2) + 8, guiwidth * 0.81, 
						lineheight * 0.8))
				.action_({arg me;
					guichild[i * 15 + 3].value_(me.lo * mixdur);
					// hmm.. this isn't working
					this.trackend(i, (me.hi - me.lo) * mixdur + guichild[i * 15 + 8].value);
					//(mixstart + (endtime - starttime)) / mixdur
					});
			};
		
		// Play button
		
		playbutton = SCButton(mixgui,
				Rect(guiwidth * 0.02, guiheight - 220, guiwidth * 0.09, lineheight))
			.states_([
				["play mix", Color.black, Color(0.3, 0.7, 0.3, 0.3)],
				["stop mix", Color.black, Color(0.7, 0.3, 0.3, 0.3)]])
			.action_({arg me;
				var actions;
				actions = [
					{this.stopmix},
					{this.playmix(starttime.value)}
					];
				actions[me.value].value
				});
		
		SCButton(mixgui,
				Rect(guiwidth * 0.12, guiheight - 220, guiwidth * 0.09, lineheight))
			.states_([
				["write nrt", Color.black, Color(0.3, 0.7, 0.3, 0.3)]]);
		
		SCButton(mixgui,
				Rect(guiwidth * 0.22, guiheight - 220, guiwidth * 0.09, lineheight))
			.states_([
				["save mix", Color.black, Color(0.3, 0.7, 0.3, 0.3)]]);
				
		SCButton(mixgui,
				Rect(guiwidth * 0.32, guiheight - 220, guiwidth * 0.09, lineheight))
			.states_([
				["load a mix", Color.black, Color(0.3, 0.7, 0.3, 0.3)]]);
				
		SCStaticText(mixgui,
				Rect(guiwidth * 0.42, guiheight - 220, guiwidth * 0.09, lineheight))
			.string_("set global output");
			
		SCPopUpMenu(mixgui, 
				Rect(guiwidth * 0.52, guiheight - 220, guiwidth * 0.09, lineheight))
			.items_([
				"track setting", "straight", "stereo", "uhj", "quad", "hex"]);
				
		SCStaticText(mixgui, 
				Rect(guiwidth * 0.02, guiheight - 216 + lineheight, guiwidth * 0.09, lineheight))
			.string_("Amp (in dB)");
			
		SCSlider(mixgui, 
				Rect(guiwidth * 0.12, guiheight - 216 + lineheight, guiwidth * 0.39, lineheight))
			.value_(0.7079402)
			.action_({arg me;
				var val;
				val = ControlSpec(-90, 6, \db).map(me.value);
				guichild[guichild.indexOf(me) + 1].value_(
					val.trunc(0.01));
				guichild[guichild.indexOf(me) + 1].bounds.left_(
					me.bounds.left)
				});
			
		SCNumberBox(mixgui, 
				Rect(guiwidth * 0.52, guiheight - 216 + lineheight, guiwidth * 0.06, lineheight))
			.value_(0.0)
			.action_({arg me;
				var val;
				val = ControlSpec(-90, 6, \db).unmap(me.value); //map from -inf to 6db
				guichild[guichild.indexOf(me) - 1].value_(val.postln);
				});
		
		SCStaticText(mixgui, 
				Rect(guiwidth * 0.02, guiheight - 212 + (lineheight * 2), guiwidth * 0.09, 
					lineheight))
			.string_("Mix Start Time =");
			
		starttime = SCNumberBox(mixgui,
				Rect(guiwidth * 0.12, guiheight - 212 + (lineheight * 2), guiwidth * 0.09, 
					lineheight))
			.value_(0.0)
			.action_({arg me;
				clock.string_(me.value.asTimeString)
				});	

		SCStaticText(mixgui, 
				Rect(guiwidth * 0.22, guiheight - 212 + (lineheight * 2), guiwidth * 0.09, 
					lineheight))
			.string_("Mix End Time =");
			
		endtime = SCNumberBox(mixgui,
				Rect(guiwidth * 0.32, guiheight - 212 + (lineheight * 2), guiwidth * 0.09, 
					lineheight))
			.value_(0.0);
			
		durstring = SCStaticText(mixgui, 
				Rect(guiwidth * 0.42, guiheight - 212 + (lineheight * 2), guiwidth * 0.19,
					lineheight))
			.string_("Mix dur = 0.0");
							
		SCStaticText(mixgui,
				Rect(guiwidth * 0.02, guiheight - 208 + (lineheight * 3), guiwidth * 0.09, 
					lineheight))
			.string_("Global Env");
			
		SCTextField(mixgui, 
				Rect(guiwidth * 0.09, guiheight - 208 + (lineheight * 3), guiwidth * 0.29, 
					lineheight))
			.value_("nil")
			.action_({arg me; 
		/*			me.value.interpret.isKindOf(Env).if({
					env = me.value.interpret;
					env.releaseNode.isNil.if({
						this.endtime_(starttime + (env.times.sum * timescale))})}, {
					}, {
					"Must be an env object".warn;
					me.value_(env.asCompileString)}) */
				});
		
		clock = SCStaticText(mixgui,
				Rect(guiwidth * 0.65, guiheight - 220, guiwidth * 0.3, lineheight * 1.8))
			.string_(0.asTimeString)
			.font_(Font("Helvetica", 30));
		
		mixgui.onClose({mixgui = nil});
		guichild = mixgui.view.children;
		
		}
	
	*initClass {
		formats = IdentityDictionary[
			\straight -> 0,
			\ster -> 1,
			\uhj ->2,
			\quad -> 3,
			\hex -> 4,
			];
		numouts = IdentityDictionary[
			\straight -> 0,
			\ster -> 2,
			\uhj ->2,
			\quad -> 4,
			\hex -> 6,
			];
		outvals = IdentityDictionary[
			\analog -> 0,
			\optical -> 8,
			\mkiianalog -> 2,
			\mkiioptical -> 14,
			];

		}
	
	}
	
/* TO DO

problems playing more then one track!  Problem with Player

All players should route out to an virtual bus, to be read back and enveloped accordingly
add add action and target vals to Player

Play func controls start (or delay) times for the Players
Player.new.outbus -> virtual bus

Mixer SynthDef to read off of it.

Make all Players read off a buffer for Envs (rather then using EnvGen). easier to start and end in the middle of the file

add a generic output function (takes speaker positions)

*/